#!/usr/bin/env python
# coding: utf8

# XXXX/TODO: use the module from pyauxm

from __future__ import print_function, unicode_literals, absolute_import, division

import os
import random
import sys
import twisted.python.log as twisted_log
from . import exc_log


class ExcLogObserver(object):
    """ A Twisted Log Observer that uses advanced logging for exceptions """
    # See twisted.python.log.PythonLoggingObserver for information on what is what

    def emit(self, eventDict):
        if not eventDict.get('isError'):
            return  # not for us
        # Obtain the info:
        vf = eventDict.get('failure')
        if vf is not None:
            e_t, e_v, e_tb = vf.type, vf.value, vf.getTracebackObject()
        else:
            e_t, e_v, e_tb = sys.exc_info()
        # sys.excepthook(e_t, e_v, e_tb)  # default-hook
        exc_log.advanced_info_safe(e_t, e_v, e_tb)  # exc_log

    def start(self):
        twisted_log.addObserver(self.emit)

    def stop(self):
        twisted_log.removeObserver(self.emit)


# A singleton instance for convenience
exc_log_observer = ExcLogObserver()


def del_twisted_default_log():
    """ Remove the default log Observer from twisted handler """
    try:
        twisted_log.defaultObserver.stop()
    except ValueError:  # probably removed already
        pass


def make_manhole_telnet(
        socket="./manhole.sock", socket_kwa=None, ns1=None, ns2=None,
        auto_ns=True, verbose=False, run=True, run_reactor=False):
    """ Creates a manhole telnet server over unix socket and returns a
    callable that adds it to the reactor (or runs it if `run`, which
    is default). It needs a running Twisted reactor to work (which is
    started and blocks if `run_reactor`).

    NOTE: authentication is not supported: use either socket file
        permissions or use SSH manhole `make_manhole` instead.

    :param socket_kwa: additional arguments for `reactor.listenUNIX`.
    :param ns1: will be available (if set) under `ns1` in the shell,
    :param ns2: dict whose contents will be available in the namespace itself.
    :param auto_ns: use full namespace (and also `self.__dict__` if
        present) of the caller.
    """
    # ... though I don't even completely know how to add auth in here anyway >.>
    from twisted.internet import reactor
    from twisted.conch import telnet, manhole
    from twisted.conch.insults import insults
    from twisted.internet import protocol
    from twisted.internet import error

    # Namespace stuff
    ns_b = {}
    if auto_ns:
        ns_b.update(sys._getframe(1).f_globals)
        caller_locals = sys._getframe(1).f_locals
        selfdict = getattr(caller_locals.get('self'), '__dict__', None)
        if isinstance(selfdict, dict):
            ns_b.update(selfdict)
        ns_b.update(caller_locals)

    def createShellServer():
        if verbose:
            print('Creating telnet server instance')
        # Namespace stuff
        ns = ns_b
        if ns1 is not None:
            ns['ns1'] = ns1
        if ns2 is not None:
            ns.update(ns2)

        factory = protocol.ServerFactory()

        class TT_LogFix(telnet.TelnetTransport):
            """ A semi-hax-fix protocol that silents out
            ConnectionDone errors """

            def connectionLost(self, reason):
                if (reason.check(error.ConnectionDone) or
                        reason.check(error.ConnectionLost)):
                    return
                return telnet.TelnetTransport.connectionLost(self, reason)

        # This does:
        # f.protocol = lambda: ProtoA(ProtoB, ProtoC, ProtoD, ProtoD_args)
        # (A instantiates B(...) who instantiates C(...) who instantiates D(ProtoD_args))
        factory.protocol = lambda: (
            TT_LogFix(  # telnet.TelnetTransport(
                telnet.TelnetBootstrapProtocol,
                insults.ServerProtocol,  # NOTE: need better terminal (likely)
                manhole.ColoredManhole,
                ns))
        if verbose:
            print('Listening on %r' % (socket,))

        lport = reactor.listenUNIX(  # pylint: disable=E1101
            socket, factory, **(socket_kwa or {}))
        return lport

    def run_manhole(run_reactor=run_reactor):
        """ RTFS """
        reactor.callWhenRunning(  # pylint: disable=E1101
            createShellServer)
        if run_reactor:
            return reactor.run()  # pylint: disable=E1101
        else:
            return reactor.run  # pylint: disable=E1101

    if run:
        return run_manhole()
    else:
        return run_manhole


def make_manhole(
        port=2000, auth_data=None,
        auth_keys_files='~/.ssh/authorized_keys_manhole', ns1=None,
        ns2=None, auto_ns=True, verbose=False, run=True,
        run_reactor=False):
    """ Creates a manhole SSH server and returns a callable that adds
    it to the reactor (or runs it if `run`, which is default). It
    needs a running Twisted reactor to work (which is started and
    blocks if `run_reactor`).

    :param port: TCP port if int or unix socket if str.

    To connect to a unix socket, use
    `ssh -o ProxyCommand='/bin/nc.openbsd -U %h' filename ...`

    :param auth_data: list or dict of (username, password);
        if password is None, it will be auto-generated and printed.
    :param auth_keys_files: list of filenames with
        'authorized_keys'-like files ('~' is expanded, relative path
        are relative to the __main__).
    :param ns1: will be available (if set) under `ns1` in the shell,
    :param ns2: dict whose contents will be available in the namespace itself.
    :param auto_ns: use full namespace (and also `self.__dict__` if
        present) of the caller.
    """
    ## Everything is cramped in this one function for simplicity. Uncramp
    ##  and put into a separate module if needs be.
    from twisted.internet import reactor
    from twisted.cred import portal, checkers
    from twisted.conch import manhole, manhole_ssh
    from twisted.conch.checkers import SSHPublicKeyDatabase
    from twisted.python.filepath import FilePath

    if isinstance(port, (bytes, str)):
        use_unix_sock = True
    elif isinstance(port, int):
        use_unix_sock = False
    else:
        raise Exception("`port` is of unknown type")

    ns_b = {}
    if auto_ns:
        ns_b.update(sys._getframe(1).f_globals)
        caller_locals = sys._getframe(1).f_locals
        selfdict = getattr(caller_locals.get('self'), '__dict__', None)
        if isinstance(selfdict, dict):
            ns_b.update(selfdict)
        ns_b.update(caller_locals)

    class AuthorizedKeysChecker(SSHPublicKeyDatabase):
        def __init__(self, authorized_keys_files):
            # # Twisted... old-style classes.
            # super(AuthorizedKeysChecker, self).__init__()
            # # And no __init__ at all anyway.
            # SSHPublicKeyDatabase.__init__(self)
            if not isinstance(authorized_keys_files, (list, tuple)):
                authorized_keys_files = [authorized_keys_files]

            # The horror? Relative path in params are relative to the __main__
            basebasepath = os.path.dirname(os.path.abspath(
                sys.modules['__main__'].__file__))

            def process_filepath(path):
                # `path.startswith('/'):` but crossplatform
                if os.path.isabs(path):
                    return path  # already absolute
                elif path.startswith('~'):
                    return os.path.expanduser(path)
                else:  # Expand relative otherwise.
                    return os.path.join(basebasepath, path)

            self.authorized_keys_files = [
                FilePath(process_filepath(v)) for v in authorized_keys_files]

        def getAuthorizedKeysFiles(self, credentials):
            # # This function is overridden from parent
            # return [FilePath(v) for v in self.authorized_keys_files]
            return self.authorized_keys_files

    def createShellServer():
        if verbose:
            print('Creating shell server instance')
        z_auth_data = dict(auth_data or {})
        for k, v in z_auth_data.items():
            if v is None:
                rndpwd = "%x" % random.getrandbits(128)
                z_auth_data[k] = rndpwd
                print("Manhole password for %r: %r" % (k, rndpwd))

        # Namespace stuff
        ns = ns_b
        if ns1 is not None:
            ns['ns1'] = ns1
        if ns2 is not None:
            ns.update(ns2)

        mhrealm = manhole_ssh.TerminalRealm()
        mhrealm.chainedProtocolFactory.protocolFactory = (
            lambda v: manhole.Manhole(ns))
        mhportal = portal.Portal(mhrealm)
        mhportal.registerChecker(
            checkers.InMemoryUsernamePasswordDatabaseDontUse(**z_auth_data))
        mhportal.registerChecker(AuthorizedKeysChecker(auth_keys_files))
        ## pubkeys from parameters... not very needed.
        #mhportal.registerChecker(PubKeysChecker(zauthorizedKeys))
        mhfactory = manhole_ssh.ConchFactory(mhportal)
        if verbose:
            print('Listening on %r' % (port,))
        if use_unix_sock:
            lport = reactor.listenUNIX(  # pylint: disable=E1101
                port, mhfactory)
        else:
            lport = reactor.listenTCP(  # pylint: disable=E1101
                port, mhfactory)
        return lport

    def run_manhole(run_reactor=run_reactor):
        """ RTFS """
        if verbose:
            print('Registering Manhole server with the reactor')
        reactor.callWhenRunning(  # pylint: disable=E1101
            createShellServer)
        if run_reactor:
            if verbose:
                print('Running Twisted Reactor')
            return reactor.run()  # pylint: disable=E1101
        else:
            return reactor.run  # pylint: disable=E1101

    if run:
        return run_manhole()
    else:
        return run_manhole



#!/usr/bin/env python

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function, unicode_literals, absolute_import, division

import os
import math
import struct
import sys
from six.moves import cPickle as pickle


__author__ = "Pedro Larroy"
__version__ = "0.1"
__date__ = "2011-07-29"
__maintainer__ = "Pedro Larroy http://pedro.larroy.com"


def numcores():
    return int(os.popen('grep core\ id /proc/cpuinfo  | wc -l').read()[:-1])


def forkfun(accumulator, function, sequence):
    # IPC stuff
    structformat = 'L'
    structlen = struct.calcsize(structformat)

    def sendmessage(myend, message):
        """ Send a pickled message across a pipe """
        outobj = pickle.dumps(message)
        return os.write(myend, struct.pack(structformat, len(outobj)) + outobj)

    def recvmessage(myend):
        """ Receive a pickled message from a pipe """
        buf = os.read(myend, structlen)
        if not buf:
            raise EOFError('EOF on pipe from parent, pid: {0}'.format(os.getpid()))

        length = struct.unpack(structformat, buf)[0]
        return pickle.loads(os.read(myend, length))

    try:
        maxchildren = function.parallel_maxchildren
    except AttributeError:
        maxchildren = numcores()

    assert function
    assert accumulator

    finished = 0

    # Spawn the worker children.  Don't create more than the number of
    # values we'll be processing.
    fromchild, toparent = os.pipe()
    children = []

    sequence_iter = sequence.__iter__()
    sequence_idx = 0

    for childnum in range(min(maxchildren, len(sequence))):
        fromparent, tochild = os.pipe()
        pid = os.fork()
        # Parent?
        if pid:
            os.close(fromparent)
            # Do some housekeeping and give the child its first assignment
            children.append({'pid': pid, 'tochild': tochild})
            cnt = sendmessage(tochild, (sequence_idx, sequence_iter.next()))
            if not cnt:
                raise EOFError('EOF on pipe to child {0}'.format(childnum))
            sequence_idx += 1

        # Child?
        else:
            os.close(tochild)

            # Keep processing values until the parent kills you
            while True:
                try:
                    message = recvmessage(fromparent)
                    if message is None:
                        sys.exit()
                    index, value = message
                    if not sendmessage(toparent, (childnum, index, function(value))):
                        sys.stderr.write('Child {0}: pipe eof\n'.format(childnum))

                except Exception as excvalue:
                    sys.stderr.write('Exception: {0}\n'.format(excvalue))

                    # # For debugging where the exception took place, enable this:
                    # raise

                    if not sendmessage(toparent, (childnum, index, excvalue)):
                        sys.stderr.write('Child {0}: pipe eof\n'.format(childnum))
                    sys.exit()

    # Keep accepting values back from the children until they've
    # all come back
    while finished < len(sequence):
        returnchild, returnindex, value = recvmessage(fromchild)
        if isinstance(value, Exception):
            raise value

        # If there are still values left to process, hand one
        # back out to the child that just finished
        if sequence_idx < len(sequence):
            sendmessage(children[returnchild]['tochild'],
                        (sequence_idx, sequence_iter.next()))
            sequence_idx += 1

        accumulator(value)
        finished += 1

    for child in children:
        sendmessage(child['tochild'], None)

    for child in children:
        os.wait()


def parallelizable(maxchildren=None, perproc=None):
    """ Mark a function as eligible for parallelized execution.  The
    function will run across a number of processes equal to
    maxchildren, perproc times the number of processors installed on
    the system, or the number of times the function needs to be run to
    process all data passed to it - whichever is least. """
    if not maxchildren:
        try:
            maxchildren = numcores()
        except ValueError:
            maxchildren = 2

    if perproc is not None:
        processors = 4  # hand-waving
        maxchildren = min(maxchildren, perproc * processors)

    def decorate(func):
        """Set the parallel_maxchildren attribute to the value
        appropriate for this function"""
        setattr(func, 'parallel_maxchildren', maxchildren)
        return func

    return decorate


if __name__ == '__main__':

    def f(x):
        y = x ** 3
        z = math.sqrt(y)
        return z
        # x = math.acos(z)

    def acc(x):
        print(x)
        pass

    forkfun(acc, f, range(100000))


# if __name__ == '__main__':
#     import doctest
#     doctest.testmod()
#
#     @parallelizable(10, perproc=4)
#     def timestwo(x, y):
#         return (x + y) * 2
#     print map(timestwo, [1, 2, 3, 4], [7, 8, 9, 10])
#
#     #@parallelizable(10)
#     @parallelizable()
#     def busybeaver(x):
#         for i in range(10000000):
#             x = x + i
#         return x
#     print map(busybeaver, range(27))



#!/usr/bin/env python
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials
#       provided with the distribution.
#
#     * Neither the name of Kirk Strauser nor the names of its
#       contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""
Very basic parallel processing support

Replaces the built-in map() function with a version that can run
across many processes simultaneously.  Attempts to replicate the
semantics of map() as precisely as possible so that it can be used in
all situations with little surprise.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import os
import signal
import struct
import sys
import traceback
from six.moves import cPickle as pickle
from six.moves import builtins


__author__ = "Kirk Strauser <kirk@strauser.com>"
__version__ = "$Rev: 1139 $"
__date__ = "$Date: 2007-05-24 10:56:44 -0500 (Thu, 24 May 2007) $"


def map(function, *sequence):
    """ map(function, sequence[, sequence, ...]) -> list

    Like the builtin map() function, but splits the workload across a
    pool of processes whenever possible.
    """

    # IPC stuff
    structformat = 'L'  #'H'  #not everything there is short enough.
    structlen = struct.calcsize(structformat)

    def sendmessage(myend, message):
        """Send a pickled message across a pipe"""
        outobj = pickle.dumps(message)
        os.write(myend, struct.pack(structformat, len(outobj)) + outobj)

    def recvmessage(myend):
        """Receive a pickled message from a pipe"""
        length = struct.unpack(structformat, (os.read(myend, structlen)))[0]
        return pickle.loads(os.read(myend, length))

    try:
        maxchildren = function.parallel_maxchildren
    except AttributeError:
        return builtins.map(function, *sequence)

    # Handle map()'s multi-sequence semantics
    if len(sequence) == 1:
        if function is None:
            return list(sequence[0])
        arglist = zip(sequence[0])
    else:
        arglist = builtins.map(None, *sequence)
    if function is None:
        return arglist

    argindex = 0
    finished = 0
    outlist = [None] * len(arglist)

    # Spawn the worker children.  Don't create more than the number of
    # values we'll be processing.
    fromchild, toparent = os.pipe()
    children = []
    for childnum in range(min(maxchildren, len(arglist))):
        fromparent, tochild = os.pipe()
        pid = os.fork()
        # Parent?
        if pid:
            # Do some housekeeping and give the child its first assignment
            children.append({
                'pid': pid,
                'fromparent': fromparent,
                'tochild': tochild,
            })
            sendmessage(tochild, (argindex, arglist[argindex]))
            argindex += 1
        # Child?
        else:
            # Since children can't really tell when they've been
            # orphaned, set a timeout so that they die if they don't
            # hear from the parent in a timely manner.
            def timeouthandler(signum, frame):
                """Get out cleanly"""
                sys.exit()
            oldsignal = signal.signal(signal.SIGALRM, timeouthandler)

            # Keep processing values until the parent kills you
            while True:
                try:
                    # Wait one second before quitting.  Children
                    # should generally hear from their parent almost
                    # instantly.
                    signal.alarm(1)
                    message = recvmessage(fromparent)
                    signal.alarm(0)
                    if message is None:
                        sys.exit()
                    index, value = message
                    sendmessage(toparent, (childnum, index, function(*value)))
                except Exception as excvalue:
                    try:
                        excvalue.tb = traceback.extract_tb(sys.exc_info()[2])
                    except Exception as e2:
                        try:
                            excvalue.e2 = e2
                        except Exception as e3:
                            pass
                    sendmessage(toparent, (childnum, index, excvalue))
                finally:
                    signal.signal(signal.SIGALRM, oldsignal)

    # Keep accepting values back from the children until they've all
    # come back
    while finished < len(arglist):
        returnchild, returnindex, value = recvmessage(fromchild)
        if isinstance(value, Exception):
            try:
                print("Child traceback:")
                print(''.join(traceback.format_list(value.tb)))
            except Exception as e2:
                pass
            raise value
        outlist[returnindex] = value
        finished += 1
        # If there are still values left to process, hand one back out
        # to the child that just finished
        if argindex < len(arglist):
            sendmessage(children[returnchild]['tochild'],
                        (argindex, arglist[argindex]))
            argindex += 1

    # Kill the child processes
    for child in children:
        sendmessage(child['tochild'], None)
    for child in children:
        os.wait()

    return outlist


def parallelizable(maxchildren=2, perproc=None):
    """ Mark a function as eligible for parallelized execution.  The
    function will run across a number of processes equal to
    maxchildren, perproc times the number of processors installed on
    the system, or the number of times the function needs to be run to
    process all data passed to it - whichever is least. """
    if perproc is not None:
        processors = 4 # hand-waving
        maxchildren = min(maxchildren, perproc * processors)

    def decorate(f):
        """Set the parallel_maxchildren attribute to the value
        appropriate for this function"""
        setattr(f, 'parallel_maxchildren', maxchildren)
        return f

    return decorate


if __name__ == '__main__':
    import doctest
    doctest.testmod()

    @parallelizable(10, perproc=4)
    def timestwo(x, y):
        return (x + y) * 2

    print(map(timestwo, [1, 2, 3, 4], [7, 8, 9, 10]))

    @parallelizable(10)
    def busybeaver(x):
        for i in range(1000000):
            x = x + i
        return x

    print(map(busybeaver, range(27)))



""" Various postgres-specific helpers, mostly dependent on django orm setup
"""


def do_sql_load_file(copy_sql, the_file, commit=True):
    if isinstance(the_file, str):
        f = open(the_file)
    else:
        f = the_file
    from django.db import connection, transaction
    cursor = connection.cursor()
    ## The very postgresql-specific stuff (also, namespace funhorror)
    cursor.cursor.cursor.copy_expert(copy_sql, f)
    if commit:
        transaction.commit_unless_managed()


def py_to_psqls(item):
    """ Convert the provided python object into a postgres-dumpfile-compatible
    string """
    ## XXX: additional type-specific processing might be requred
    if item == None:
        return 'NULL'
    return str(item)


class ModelToDump(object):
    """ A helper class to convert a django orm Model into a sql / data file
    for loading into the database """

    tfile = None

    def __init__(self, model, the_file=None, write_sql=False):
        self.model = model
        self.write_sql = write_sql
        mm = self.model._meta
        self.aclist = [
            f.get_attname_column()
            for f in mm.fields if not f.primary_key]
        self.attrlist = [attname for attname, column in self.aclist]
        if isinstance(the_file, str):
            self.tfile = open(the_file, 'w')
        else:  # isinstance(the_file, file):
            self.tfile = the_file
        if write_sql:
            self.tfile.write(self.get_copy_sql())

    def get_copy_sql(self):
        """ Returns the corresponding SQL `COPY` statement for the model """
        mm = self.model._meta
        # TODO: use '\\N' for that purpose.
        return (
            "COPY %s (%s) FROM stdin WITH DELIMITER AS '\t' NULL AS"
            " 'NULL';") % (
                mm.db_table,
                ', '.join('"%s"' % (cn,) for an, cn in self.aclist))

    def get_dumpline(self, minstance):
        """ Returns a line with the data from the model instance """
        vals = (getattr(minstance, an, None) for an in self.attrlist)
        return '\t'.join(py_to_psqls(val) for val in vals)

    def write_instance(self, minstance):
        """ Writes the instance to the associated file (assuming that the
        file was provided) """
        return self.tfile.write(self.get_dumpline(minstance) + "\n")

    def finalize(self):
        """ Close the file, writing ending sequence if necessary """
        if self.write_sql:
            self.tfile.write('\\.\n')
        self.tfile.close()
